
// braucht ws2_32.lib

// Anzahl der Prozesse 4

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


/* Parallele Multiplikation eines Vektors mit einem Faktor */

// Erzeugt einen zweidimensionales Gitter und 
// bildet auf dieser Topologie eine globale Summe

/*
Beispiel: 3x4
(-1,-1)	(-1,0)	(-1,1)	(-1,2)	(-1,3)	(-1,4)
(0,-1)	(0,0)	(0,1)	(0,2)	(0,3)	(0,4)
(1,-1)	(1,0)	(1,1)	(1,2)	(1,3)	(1,4)
(2,-1)	(2,0)	(2,1)	(2,2)	(2,3)	(2,4)
(3,-1)	(3,0)	(3,1)	(3,2)	(3,3)	(3,4)

Beispiel: 3x2
(-1,-1)	(-1,0)	(-1,1)	(-1,2)	
(0,-1)	(0,0)	(0,1)	(0,2)
(1,-1)	(1,0)	(1,1)	(1,2)
(2,-1)	(2,0)	(2,1)	(2,2)
(3,-1)	(3,0)	(3,1)	(3,2)


Beispiel: 2x3, erst Zeilen, dann Spalten

		(0,0)	(0,1)	(0,2)	
		(1,0)	(1,1)	(1,2)	
		

(-1,-1)	(-1,0)	(-1,1)	(-1,2)	(-1,3)
(0,-1)	(0,0)	(0,1)	(0,2)	(0,3)
(1,-1)	(1,0)	(1,1)	(1,2)	(1,3)
(2,-1)	(2,0)	(2,1)	(2,2)	(2,3)

*/



#include <windows.h>
//#include <sys/param.h>

#define	ROOT		0
#define	TAG			1
#define	UP			1
#define DOWN		-1


// Prototyp
char*	PrintTopologieType(int, char*);



int main(int argc, char *argv[]) {

	MPI_Status	status;
	int 		Pid, nTasks;
	int			i;
	int			nDims;	
	int			MpiDims[2] = {2,2};
	int			Periods[2] = {FALSE, FALSE};

/*
periods - logisches Array der Gre ndims (array of integer)
periods[n]=true : Struktur ist in der Dimension n periodisch
periods[n]=false: Struktur ist in der Dimension n nicht periodisch 
   Zyklus innerhalb einer Dimension  -1 wird auf n-1 gemappt
*/
	int			Recorder= FALSE; // Bei True => Umndern der IDs 
	MPI_Comm	newComm;  // definiert die neue Welt, Matrix

	int			TopoType; // fr Test-Ausgabe
	char 		myString[80];	 // fr Test-Ausgabe

	char		hostname[260];

	int			Coord[2];  // speichert die Koordinaten pro Task
	int			x,y;
	int			*rDims, *rPeriods, *rCoords;

	int			prev, next;
/*	
	int			SourceRank, DestRank;
	int 		TranslateRank;
	int			Neighbors[2];
	int			Send, Recv, Sum, TopoType;
*/

	int retcode;
	int t1, t2, temp;

	int nr, Value, MaxValue;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &Pid);
	MPI_Comm_size(MPI_COMM_WORLD, &nTasks);

	if ( (   (nTasks%2) == 1) ){
      puts("Bitte genau mit einer geraden Anzahl von Prozessen starten");
	  return 1;
	}
	if ( nTasks < 4 ){
      puts("Bitte mit mindestens vier Prozessen starten");
	  return 1;
	}
  
	Value = Pid+100;
	
	gethostname(hostname, 258);

	// Benutzt die Hilfsfunktion Dims_Create
	// erzeugt im Feld MpiDims die Zahlen, wenn Inhalt = 0
	nDims = 2;
	MpiDims[0] = 2;  // zwei Zeilen
	MpiDims[1] = 0;  // drei Spalten
	retcode = MPI_Dims_create(nTasks, nDims, MpiDims);
	if (retcode != MPI_SUCCESS) {
		puts("Fehler bei DIMS_Create");
	}
	//printf("Ergebnis MPI_Dims_create: %d\n",retcode);

	if (Pid == ROOT) {
		printf("Vorschlag von MPI fuer %d Knoten und %d Dimensionen : ",
		       nTasks, nDims);
		for (i = 0; i < nDims; i++)
			printf("%d ", MpiDims[i]);
		printf("\n");
	}
	// manuelle nderung
	//MpiDims[0] = 1;  // zwei Zeilen
	//MpiDims[1] = 4;  // drei Spalten

	// Nun hat jeder Prozess eine Pid  UND eine Koordinate (i,j)
	
	// Erzeugen der neuen Topologie, Ergebnis in newComm
	Recorder=FALSE; // MPI kann die IDs verndern, besseres Laufzeitverhalten
	Periods[0] = FALSE;
	Periods[1] = FALSE;
	retcode = MPI_Cart_create(MPI_COMM_WORLD, nDims, MpiDims, Periods, Recorder, &newComm);
	if (retcode != MPI_SUCCESS) {
		puts("Fehler bei MPI_Cart_create");
	}
	//printf("Ergebnis MPI_Cart_create: %d\n",retcode);

	printf("Pid: %d   Ergebnis newComm: %d\n",Pid, newComm);
	
	// Ergebnis ausgeben
	
	if (Pid == ROOT) {
		MPI_Topo_test(newComm, &TopoType);
		printf ("Topology type ist %s\n", PrintTopologieType(TopoType, myString));
	}
	

    if (TRUE) {
		retcode = MPI_Cartdim_get(newComm, &nDims);
		if (retcode != MPI_SUCCESS) {
			puts("Fehler bei Cartdim_get");
		}
		printf("Ergebnis MPI_Cartdim_get: %d  Anz-Dimensionen: %d\n",retcode, nDims);
	}

	// holen der Koordinaten

	rDims 	 = (int*) malloc (nDims * sizeof(int));
	rPeriods = (int*) malloc (nDims * sizeof(int));
	rCoords  = (int*) malloc (nDims * sizeof(int));

	retcode = MPI_Cart_get(
		  newComm, nDims, rDims, rPeriods, rCoords);
	if (retcode != MPI_SUCCESS) {
		puts("Fehler bei MPI_Cart_create");
	}


/*	printf("%-15.12s: Topology Informationen : Dims = %d\n", hostname, nDims);
	for (i = 0; i < nDims ; i++) 
		printf("%-15.12s: Processes in Dim[%d] = %d  Periodicity = %d  my Coordinates = %d\n",
        hostname, i, rDims[i], rPeriods[i], rCoords[i]);
*/

	//liefert aus der Pid, Rank die Koordinaten

	MPI_Cart_coords(newComm, Pid, nDims, Coord); 
	x = Coord[1];
	y = Coord[0];
	printf("Pid: %d  y: %d  x:%d   DIM y: %d  x:%d\n",Pid,y,x, rCoords[0], rCoords[1]);


	// ------------------------------
	// bestimme Vorgnger und Nachfolger
	/*
		if (Pid == ROOT) {
		prev=1000;
		next=1000;
		for (i=0; i<4; i++) {
			Coord[0]=1;
			Coord[1]=i;
			MPI_Cart_rank(newComm, Coord, &prev); 
			printf("Coord[0]: %d   Coord[1]: %d   Rank:%d\n",Coord[0], Coord[1], prev);
		}
	}*/

	prev=1000;
	next=1000;
	temp = Coord[1];
	Coord[1] --;
	if (Coord[1]<0)  Coord[1]=MpiDims[1]-1;
	t1=Coord[1];
	MPI_Cart_rank(newComm, Coord, &prev); 
	printf("Pid: %d   Coord[1]: %d Coord[0]: %d   prev: %d\n",
		  Pid,Coord[0], Coord[1], prev);

	Coord[1]=temp;
	Coord[1] ++;
	if (Coord[1]>=MpiDims[1])  Coord[1]=0;
	t2=Coord[1];
	MPI_Cart_rank(newComm, Coord, &next); 
	Coord[1]=temp;
	printf("Pid: %d   Coord[0]: %d Coord[1]: %d   next: %d\n",
		  Pid,Coord[0], Coord[1], next);


	nr=1;
	switch (nr) {
		case 0:
			// Nun die Aktion, bestimme Zeilenweise das Maximum
			if ( x==0 ) {
				MPI_Send( &Value, 1, MPI_INT, next, 0, newComm);
				MPI_Recv( &MaxValue, 1, MPI_INT, prev, 0, newComm, &status);
				printf("Pid: %d   MaxValue %d\n", Pid, MaxValue);
			}
			else {
				MPI_Recv( &MaxValue, 1, MPI_INT, prev, 0, newComm, &status);
				printf("Pid: %d   MaxValue %d\n", Pid, MaxValue);
				if (Value>MaxValue) MaxValue=Value;
				MPI_Send( &MaxValue, 1, MPI_INT, next, 0, newComm);
			}
			break;
		case 1:
			// Nun die Aktion, bestimme Zeilenweise die Summe
			if ( x==0 ) {
				MPI_Send( &Value, 1, MPI_INT, next, 0, newComm);
				MPI_Recv( &MaxValue, 1, MPI_INT, prev, 0, newComm, &status);
				printf("Pid: %d   Summe %d\n", Pid, MaxValue);
			}
			else {
				MPI_Recv( &MaxValue, 1, MPI_INT, prev, 0, newComm, &status);
				printf("Pid: %d   MaxValue %d\n", Pid, MaxValue);
				MaxValue+=Value;
				MPI_Send( &MaxValue, 1, MPI_INT, next, 0, newComm);
			}
			break;
		case 2:
			// Nun die Aktion, bestimme Zeilenweise die Summe der Quadrate
			// Vektorprodukt
			if ( x==0 ) {
				MaxValue=Value*Value;
				MPI_Send( &MaxValue, 1, MPI_INT, next, 0, newComm);
				MPI_Recv( &MaxValue, 1, MPI_INT, prev, 0, newComm, &status);
				printf("Pid: %d   Summe %d\n", Pid, MaxValue);
			}
			else {
				MPI_Recv( &MaxValue, 1, MPI_INT, prev, 0, newComm, &status);
				printf("Pid: %d   MaxValue %d\n", Pid, MaxValue);
				MaxValue+=Value*Value;
				MPI_Send( &MaxValue, 1, MPI_INT, next, 0, newComm);
			}
			break;
	}


	MPI_Finalize();
	return 0;
}

char* PrintTopologieType(int tt, char *str)
{
	if (str != NULL)
	    switch (tt) {
		case MPI_UNDEFINED : strcpy(str, "MPI_UNDEFINED"); break;
		case MPI_GRAPH	   : strcpy(str, "MPI_GRAPH"); break;
		case MPI_CART	   : strcpy(str, "MPI_CART"); break;
	    }
	return str;
}
