#include "mpi.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>


/*
 Version 3
 Aufteilen von 1 bis n auf     m Workers 
 Die Aufteilung wird ber ein zustzliches Feld realisiert
    siehe Feld    offset  und sendcounts
   MPI_Scatterv
 Nach der bermittlung wird eine Funktion ausgefhrt, add um 2
 und dann werden mittels der Gather-Anweisung die Daten wieder gesammelt
 MPI_Gatherv
*/

#define C_MAX 12	

#define ROOT 0

// Scatter 3

int main(int argc, char ** argv) {
	int     root = 0;
	int     numprocs, Pid, i;
	int     x_root[C_MAX+1];
	int     y_root[C_MAX+1];
	int     client_size;
    

	int sendcounts[C_MAX+1]; // speichert die Lngen 
	int offset[C_MAX+1]; // speichert die Lngen
	int rest;
	int summe;
	int retcode;

	int     x[C_MAX+1];
	int     y[C_MAX+1];

	if (MPI_Init(&argc,&argv)!=MPI_SUCCESS) {
		printf("MPI_Init failed.\n");
		return -1;
	}

	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &Pid);

	printf("Process-Id:  %d: Anzahl der Prozesse: %d.\n", Pid, numprocs);


	if (numprocs>C_MAX) {
		puts("Fehler, zu viele Prozesse");
	}

	// Master baut die Daten auf
	 if (Pid==0) {
		  // Daten im Master erzeugen
		for (i=0; i<C_MAX; i++) {
			x_root[i] = i+1;
		}
	}

	  // zur Kontrolle werden die lokalen Felder auf Null gesetzt, pro Client
	for (i=0; i<C_MAX; i++) {
		x[i] = 0;
	}

	   // Berechnen der Aufteilung des Feldes
	client_size = C_MAX / numprocs;  // ganzzalig !!!
	for (i=0; i<C_MAX; i++) {
		sendcounts[ i ] = client_size;
	}

	  // gib es ein Rest ?
	rest = C_MAX  - numprocs*client_size;
	if (rest>0) {
		// es fehlen noch welche,  verteilen vom Anfang
		for (i=0; i<rest; i++) {
			sendcounts[ i ] += 1;
		}
	}

	 // Berechne den StartOffset pro Client
	summe=0;
	for (i=0; i<C_MAX; i++) {
		offset[ i ] = summe;
		summe += sendcounts[ i ];
	}
	

	if (Pid==0) {
		// Ausgabe der Bereiche, nur Master
		for (i=0; i<numprocs; i++) {
			printf("Ausgabe Root: i:%d  offset %d  sendcounts:  %d  \n", 
						i, offset[i], sendcounts[i]);		
		}
		puts("  ");
		for (i=0; i<C_MAX; i++) {
			printf("Ausgabe Root:   i:  %d  x_root[i]: %d\n",i, x_root[i]);
		}
		puts("  ");
	}

	// Verteilen des Feldes
	retcode = MPI_Scatterv(x_root, sendcounts, offset, MPI_INT,     
			x, sendcounts[Pid],  MPI_INT, ROOT, MPI_COMM_WORLD);
	
	if ( retcode!= MPI_SUCCESS) {
		puts("Fehler beim Aufruf von Scatter");
        		MPI_Finalize();
		return 1;
	}
  
	puts("  ");
	puts("Ausgabe der bermittelten Clientdaten:");

	  // zur Kontrolle ausgeben
	for (i=0; i<sendcounts[Pid]; i++) {
		printf(" Recv-Data:   Pid %d  i:  %d  x[i]: %d\n",Pid, i, x[i]);		
	}

	 // Berechnen der Funktionen     y=x+2, gilt auch fr Root
	for (i=0; i<sendcounts[Pid]; i++) {
		printf("Calc:  Pid %d  i:  %d  x[i]: %d\n",Pid, i, x[i]);		
		y[i] = x[i] + 2;
	}

	// Sammeln der Daten mit Gatherv an den Root
	retcode = MPI_Gatherv(y, sendcounts[Pid], MPI_INT,  
		y_root, sendcounts,  offset, MPI_INT, ROOT, MPI_COMM_WORLD);
	
	if ( retcode!= MPI_SUCCESS) {
		puts("Fehler beim Aufruf von Gather");
	}

	 // Ausgabe der Ergebnisse
	if (Pid == 0 ) {
		puts("  ");
		puts("Root:  Ausgabe der Ergebnisse:");
		for (i=0; i<C_MAX; i++) {
			printf("Root:  i:  %d  y_root[i]: %d\n", i, y_root[i]);
		}
		puts("  ");
	}


    MPI_Finalize();
	return 0;
}  // main


